@@ -2,6 +2,7 @@ from __future__ import absolute_import
 
 import json
 
+import bottle
 from bottle import request, response
 
 from .collections import merge
@@ -44,7 +45,11 @@ class JsonBottle(object):
     """Function parses input from JSON body or GET parameters"""
     def decorator(*args, **kwargs):
       # get args determined by JSON input
-      kwargs.update(json_args(request))
+      try:
+        kwargs.update(json_args(request))
+      except Exception as e:
+        request.body.seek(0)
+        raise JsonInputException("Unable to parse JSON arguments: {}".format(request.body.read()))
       return decorated(*args, **kwargs)
     return decorator
 
@@ -56,9 +61,11 @@ class JsonBottle(object):
 
       # return it as a string
       response.content_type = "application/json"
-      response.body = json.dumps(output)
-
-      return response
+      try:
+        response.body = json.dumps(output)
+        return response
+      except (ValueError, TypeError) as e:
+        raise JsonOutputException("Unable to encode output as JSON: {}".format(output))
 
     return decorator
 
@@ -90,10 +97,39 @@ class JsonBottle(object):
       f = self.json_output(f)
       f = self.json_input(f)
       f = self.cors(f)
+      f = self.on_exception(Exception)(f)
       f = self.route(*args, **kwargs)(f)
       return f
     return decorator_
 
+  def on_exception(self, exceptions=Exception, callback=None):
+    """Catch exceptions thrown by decorated function and respond with a callback
+
+    Parameters
+    ----------
+    exceptions : tuple or Exception
+        exceptions to catch
+    callback : None or function
+        Callback applied when an exception is thrown. If callback is None, a
+        dict with the exception's details is returned; otherwise, the
+        callback's output is returned.
+    """
+
+    def decorator_(f):
+      def decorator(*args_, **kwargs_):
+        try:
+          return f(*args_, **kwargs_)
+        except exceptions as e:
+          if callback is None:
+            return {
+                "status"    : "exception",
+                "exception" : e.__class__.__name__ + ": " + str(e)
+              }
+          else:
+            return callback(e)
+      return decorator
+    return decorator_
+
 
 def json_args(r):
   """Get JSON arguments to this request
@@ -116,17 +152,17 @@ def json_args(r):
   kwargs : dict
       collected keyword arguments from request
   """
-  # from POSTed content
-  if r.json is not None:
-    return r.json
+  r.body.seek(0)
+  content = r.body.read()
+  if content:
+    # body is assumed to be JSON-encoded content
+    return json.loads(content)
+  elif r.method.upper() == "GET":
+    # GET parameters might be specified
+    return dict(r.params)
   else:
-    # from body
-    try:
-      r.body.seek(0)
-      return json.loads(r.body.read())
-    except ValueError:
-      # from params dict
-      return dict(r.params)
+    # no parameters here
+    return {}
 
 
 def cors_headers(r):
@@ -157,3 +193,11 @@ def cors_headers(r):
       'Origin, Accept, Content-Type, X-Requested-With'
   )
   return headers
+
+
+class JsonInputException(bottle.BottleException):
+  pass
+
+
+class JsonOutputException(bottle.BottleException):
+  pass
